//
//                  Purger Unit Version 1.0.0.0
//                     Written By The Death
//       thedeadh@netvision.net.il   http://security.guides.co.il
//
//
//  The purger unit gives the programmer tools to securely destroy data
//  in a way that will prevent it's reconstruction and therefore it's exposure.
//  The supplied purging methods meet the standards (NCSG-TG-025) of the American
//  DoD, Which means that the data could not be reconstructed using the most
//  powerful hardware (as for the year 1990).
//
//  The TPurger class is the base for all other purging objects.
//
//  The TStreamPurger class is a tool for purging streams. It can be used amoung
//  every decended class of TStream which doesn't hide the WriteBuffer procedure
//
//  The TFilePurger class is a tool for purging files. It can be used among
//  files to which the current user has write-access.
//
//  The TGlobal class is a tool for purging global memory objects.
//
//  The TClipboardPurger class is a tool for purging the clipboard, or a
//  TClipboard object.
//
//
//  This file may be distributed freely. If you modify this file (And you might
//  just wish to, as it might not be optimized or just good), you may distribute
//  it only if you will credit me for the initial work. In any case, this header
//  may not be removed.
//
//  This file uses the Mersenne Twister RNG. The MTRandom unit is needed to
//  compile this unit (The MTRandom unit is attached here in the Zip file).
//  I thank Makoto Matsumoto from Kyoto University for allowing me to use and
//  distribute his work among mine.
//
//  This file was written under Delphi 6.
//
//

Unit Purger;

interface

uses
  Windows, Messages, SysUtils, Classes, MTRandom, Variants, Clipbrd;

type
  TPurgingRec = Record
    Position : DWORD;
    Phase : DWORD;
    Length : DWORD;
    Buffer : Array[0..7] of Char;
    System : Array[0..3] of DWORD;   //For potential new purging methods
    Finished : Boolean;
  End;
  PPurgingRec = ^TPurgingRec;
  TPurgingMethod = Procedure(PurgeRec : PPurgingRec) of Object;
  TPurgingMethods = (pmDoD, pmSchneier);

  TPurger = class(TObject)
  private
    MTR : TMTID;
    FDeleteAfterPurge : Boolean;
    FPurgingMethod : TPurgingMethods;
    FOnPurgeStart : TNotifyEvent;
    FOnPurgeEnd : TNotifyEvent;
    FOnMassPurgeStart : TNotifyEvent;
    FOnMassPurgeEnd : TNotifyEvent;
  protected
    Purging : Boolean;
    PurgingProcedure : TPurgingMethod;
    Procedure DoD_Purge(PurgeRec : PPurgingRec);
    Procedure Schneier_Purge(PurgeRec : PPurgingRec);
    Procedure SetPurgeProcedure;
  public
    Procedure Purge; Virtual; Abstract;
    Procedure MassPurge; Virtual; Abstract;
    Procedure AddToList(Const NewItem : Pointer); Virtual; Abstract;
    Procedure DeleteFromList(Const ExstItem : Pointer); Virtual; Abstract;
    Procedure ClearList; Virtual; Abstract;
    Constructor Create; Virtual; Abstract;
    Property PurgingMethod : TPurgingMethods Read FPurgingMethod Write FPurgingMethod Default pmDoD;
    Property OnPurgeStart : TNotifyEvent Read FOnPurgeStart Write FOnPurgeStart;
    Property OnPurgeEnd : TNotifyEvent Read FOnPurgeEnd Write FOnPurgeEnd;
    Property OnMassPurgeStart : TNotifyEvent Read FOnMassPurgeStart Write FOnMassPurgeStart;
    Property OnMassPurgeEnd : TNotifyEvent Read FOnMassPurgeEnd Write FOnMassPurgeEnd;
    Property DeleteAfterPurge : Boolean Read FDeleteAfterPurge Write FDeleteAfterPurge Default False;
  end;

  TStreamPurger = Class(TPurger)
  Private
    StreamList : TList;
  Protected
  Public
    Procedure Purge(Stream : TStream); Reintroduce;
    Procedure MassPurge; Override;
    Procedure AddToList(Const NewItem : Pointer); Override;
    Procedure DeleteFromList(Const ExstItem : Pointer); Override;
    Procedure ClearList; Override;
    Procedure LoadList(List : TList);
    Constructor Create; Override;
    Destructor Destroy; Override;
  End;

  TFilePurger = Class(TPurger)
  Private
    FileList : TStringList;
    FDeleteFilePointer : Boolean;
  Public
    Procedure Purge(sFile : String); Reintroduce;
    Procedure MassPurge; Override;
    Procedure AddToList(Const NewItem : String); Reintroduce;
    Procedure DeleteFromList(Const ExstItem : String); Reintroduce;
    Procedure ClearList; Override;
    Procedure LoadList(List : TStringList);
    Property DeleteFilePointer : Boolean Read FDeleteFilePointer Write FDeleteFilePointer Default False;
    Constructor Create; Override;
    Destructor Destroy; Override;
  End;

  PClipboard = ^TClipboard;
  TClipboardPurger = Class(TPurger)
  Private
    ClipboardList : TList;
  Public
    Procedure Purge(aClipboard : PClipboard); Reintroduce;
    Procedure MassPurge; Override;
    Procedure AddToList(Const NewItem : Pointer); Override;
    Procedure DeleteFromList(Const ExstItem : Pointer); Override;
    Procedure ClearList; Override;
    Procedure LoadList(List : TList);
    Procedure PurgeTheClipboard;
    Constructor Create; Override;
    Destructor Destroy; Override;
  End;

  TGlobalPurger = Class(TPurger)
  Private
    GlobalList : TList;
  Public
    Procedure Purge(Mem : Pointer); Reintroduce;
    Procedure MassPurge; Override;
    Procedure AddToList(Const NewItem : Pointer); Reintroduce;
    Procedure DeleteFromList(Const ExstItem : Pointer); Reintroduce;
    Procedure ClearList; Override;
    Procedure LoadList(List : TList);
    Constructor Create; Override;
    Destructor Destroy; Override;
  End;

  TLocalPurger = Class(TPurger)
  Private
    LocalList : TList;
  Public
    Procedure Purge(Mem : Pointer); Reintroduce;
    Procedure MassPurge; Override;
    Procedure AddToList(Const NewItem : Pointer); Reintroduce;
    Procedure DeleteFromList(Const ExstItem : Pointer); Reintroduce;
    Procedure ClearList; Override;
    Procedure LoadList(List : TList);
    Constructor Create; Override;
    Destructor Destroy; Override;
  End;

implementation

function LocalHandle(P : Pointer) : THandle; external kernel32 name 'LocalHandle';

{ TPurger }

procedure TPurger.DoD_Purge(PurgeRec : PPurgingRec);
Var
  B, i : Byte;
begin
  B := Byte(PurgeRec^.System[0]);
  If PurgeRec^.Phase = 0 Then
  Begin
    PurgeRec^.Position := 0;
    PurgeRec^.Phase := 1;
    Randomize;
    B := Random(256) XOR Random(256) XOR Random(256);
    PurgeRec^.Phase := 1;
  End;

  For i := 0 to 7 Do
    Move(B,PurgeRec^.Buffer[i],1);
  Inc(PurgeRec^.Position,8);
  If (PurgeRec^.Position) >= PurgeRec^.Length Then
  Begin
    Case PurgeRec^.Phase Of
      1: B := Not B;
      2: B := Random(256) XOR Random(256) XOR Random(256);
      3: PurgeRec^.Finished := True;
    End;
    Inc(PurgeRec^.Phase);
  End;
  PurgeRec^.System[0] := B;
end;

Procedure TPurger.Schneier_Purge(PurgeRec : PPurgingRec);
Var
  D1, D2 : DWORD;
Begin
  If (PurgeRec^.Position = 0) And (PurgeRec^.Phase > 2)  Then
  Begin
    Randomize;
    MT_RandInit(MTR,Random(High(Cardinal)));
    MT_RandUpdate(MTR,IntToStr(MT_TimeStamp));
    MT_RandUpdate(MTR,IntToStr(Random(1000000))+IntToStr(Random(100000)));
    MT_RandUpdate(MTR,IntToStr(Random(1000000))+IntToStr(Random(100000)));
    MT_RandUpdate(MTR,IntToStr(Random(1000000))+IntToStr(Random(100000)));
    MT_RandUpdate(MTR,IntToStr(Random(1000000))+IntToStr(Random(100000)));
    MT_RandUpdate(MTR,IntToStr(Random(1000000))+IntToStr(Random(100000)));
    MT_RandUpdate(MTR,IntToStr(MT_TimeStamp));
  End;

  If PurgeRec^.Phase > 2 Then
  Begin
    D1 := MT_SecureRandNext(MTR);
    D2 := MT_SecureRandNext(MTR);
  End
  Else
  Begin
    D1 := 0;
    If PurgeRec^.Phase = 2 Then
      D1 := Not D1;
    D2 := D1;
  End;

  Move(D1,PurgeRec^.Buffer[0],4);
  Move(D2,PurgeRec^.Buffer[4],4);
  Inc(PurgeRec^.Position,8);
  If (PurgeRec^.Position >= PurgeRec^.Length) Then
  Begin
    Case PurgeRec^.Phase Of
      7 : PurgeRec^.Finished := True;
    End;
    Inc(PurgeRec^.Phase);
  End;
End;

procedure TPurger.SetPurgeProcedure;
begin
  Case Self.PurgingMethod Of
    pmDoD : PurgingProcedure := DoD_Purge;
    pmSchneier : PurgingProcedure := Schneier_Purge;
  End;
end;

{ TStreamPurger }

procedure TStreamPurger.AddToList(Const NewItem : Pointer);
begin
  inherited;
  If TObject(NewItem^).ClassType = TStream Then
  Begin
    StreamList.Add(NewItem);
    StreamList.Capacity := StreamList.Count;
  End
  Else
    Raise EInvalidCast.Create('Pointer does not point to a TStream or a TStream decended class');
end;

procedure TStreamPurger.ClearList;
begin
  inherited;
  StreamList.Clear;
  StreamList.Capacity := 0;
end;

constructor TStreamPurger.Create;
begin
  inherited;
  StreamList := TList.Create;
end;

procedure TStreamPurger.DeleteFromList(Const ExstItem : Pointer);
begin
  inherited;
  If StreamList.IndexOf(ExstItem) = -1 Then
    Raise EListError.Create('Pointer was not found on the list')
  Else
  Begin
    StreamList.Delete(StreamList.IndexOf(ExstItem));
    StreamList.Capacity := StreamList.Count;
  End;
end;

destructor TStreamPurger.Destroy;
begin
  StreamList.Free;
  inherited;
end;

procedure TStreamPurger.LoadList(List: TList);
begin
  StreamList := List;
end;

procedure TStreamPurger.MassPurge;
Var
  D : DWORD;
begin
  inherited;
  If Assigned(OnMassPurgeStart) Then
    OnMassPurgeStart(Self);
  For D := 0 To StreamList.Count -1 Do
  Begin
    Purge(TObject(StreamList[D]^) As TStream);
    If FDeleteAfterPurge Then
    Begin
      DeleteFromList(StreamList[D]);
      StreamList.Capacity := StreamList.Count;
    End;
  End;
  If Assigned(OnMassPurgeEnd) Then
    OnMassPurgeEnd(Self);
end;

procedure TStreamPurger.Purge(Stream: TStream);
Var
  Prg : TPurgingRec;
begin
  inherited;
  If Assigned(OnPurgeStart) Then
    OnPurgeStart(Self);
  Purging := True;
  SetPurgeProcedure;
  Prg.Position := 0;
  Prg.Phase := 0;
  Prg.Length := Stream.Size;
  Prg.Finished := False;
  Stream.Position := 0;
  Repeat
    PurgingProcedure(@Prg);
    If Prg.Position >= Prg.Length Then
    Begin
      Stream.WriteBuffer(Prg.Buffer, (8 -(Prg.Position - Prg.Length)) Mod 8);
      Prg.Position := 0;
      Stream.Position := 0;
    End
    Else
      Stream.WriteBuffer(Prg.Buffer, 8);
  Until Prg.Finished;
  Purging := False;
  If Assigned(OnPurgeEnd) Then
    OnPurgeEnd(Self);
end;

{ TFilePurger }

procedure TFilePurger.AddToList(const NewItem: String);
begin
  If FileExists(UpperCase(NewItem)) Then
  Begin
    FileList.Add(UpperCase(NewItem));
    FileList.Capacity := FileList.Count;
  End
  Else
    Raise EInOutError.Create('File Not Found');
end;

procedure TFilePurger.ClearList;
begin
  inherited;
  FileList.Clear;
  FileList.Capacity := 0;
end;

constructor TFilePurger.Create;
begin
  inherited;
  FileList := TStringList.Create;
end;

procedure TFilePurger.DeleteFromList(const ExstItem: String);
begin
  If FileList.IndexOf(UpperCase(ExstItem)) = -1 Then
    Raise EListError.Create('String was not found on list')
  Else
  Begin
    FileList.Delete(FileList.IndexOf(UpperCase(ExstItem)));
    FileList.Capacity := FileList.Count;
  End;
end;

destructor TFilePurger.Destroy;
begin
  FileList.Free;
  inherited;
end;

procedure TFilePurger.LoadList(List: TStringList);
begin
  FileList := List;
end;

procedure TFilePurger.MassPurge;
Var
  D : DWORD;
begin
  inherited;
  If Assigned(OnMassPurgeStart) Then
    OnMassPurgeStart(Self);
  For D := 0 To FileList.Count -1 Do
  Begin
    Purge(FileList[D]);
    If FDeleteAfterPurge Then
    Begin
      DeleteFromList(FileList[D]);
    End;
  End;
  If Assigned(OnMassPurgeEnd) Then
    OnMassPurgeEnd(Self);
end;

procedure TFilePurger.Purge(sFile: String);
Var
  Prg : TPurgingRec;
  h : THandle;
  D : DWORD;
begin
  inherited;
  If Assigned(OnPurgeStart) Then
    OnPurgeStart(Self);
  Purging := True;
  SetPurgeProcedure;
  Prg.Position := 0;
  Prg.Phase := 0;
  h := CreateFile(PChar(sFile),GENERIC_WRITE,0,Nil,OPEN_EXISTING,GetFileAttributes(PChar(sFile)),0);
  Prg.Length := GetFileSize(h,@D);
  If D > 0 Then
    Raise EWriteError.Create('File exceeds 4294967296 bytes, therefore cannot be fully purged');
  Prg.Finished := False;
  Repeat
    PurgingProcedure(@Prg);
    If Prg.Position >= Prg.Length Then
    Begin
      WriteFile(h,Prg.Buffer,(8 -(Prg.Position - Prg.Length)) Mod 8,D,Nil);
      Prg.Position := 0;
      SetFilePointer(h,0,Nil,FILE_BEGIN);
    End
    Else
      WriteFile(h,Prg.Buffer,8,D,Nil);
  Until Prg.Finished;
  If FDeleteFilePointer Then
    DeleteFile(PChar(sFile));
  CloseHandle(h);
  Purging := False;
  If Assigned(OnPurgeEnd) Then
    OnPurgeEnd(Self);
end;

{ TClipboardPurger }

procedure TClipboardPurger.AddToList(const NewItem: Pointer);
begin
  inherited;
  If TObject(NewItem^).ClassType = TClipboard Then
  Begin
    ClipboardList.Add(NewItem);
    ClipboardList.Capacity := ClipboardList.Count;
  End
  Else
    Raise EInvalidCast.Create('Pointer does not point to a TClipboard object');
end;

procedure TClipboardPurger.ClearList;
begin
  inherited;
  ClipboardList.Clear;
  ClipboardList.Capacity := ClipboardList.Count;
end;

constructor TClipboardPurger.Create;
begin
  inherited;
  ClipboardList := TList.Create;
end;

procedure TClipboardPurger.DeleteFromList(const ExstItem: Pointer);
begin
  inherited;
  If ClipboardList.IndexOf(ExstItem) = -1 Then
    Raise EListError.Create('Item Was Not Found On The List')
  Else
  Begin
    ClipboardList.Delete(ClipboardList.IndexOf(ExstItem));
    ClipboardList.Capacity := ClipboardList.Count;
  End;
end;

destructor TClipboardPurger.Destroy;
begin
  ClipboardList.Free;
  inherited;
end;

procedure TClipboardPurger.LoadList(List: TList);
begin
  ClipboardList := List;
end;

procedure TClipboardPurger.MassPurge;
Var
  D : DWORD;
begin
  inherited;
  If Assigned(OnMassPurgeStart) Then
    OnMassPurgeStart(Self);
  For D := 0 To ClipboardList.Count -1 Do
  Begin
    Purge(ClipboardList[D]);
    If FDeleteAfterPurge Then
    Begin
      DeleteFromList(ClipboardList[D]);
    End;
  End;
  If Assigned(OnMassPurgeEnd) Then
    OnMassPurgeEnd(Self);
End;

procedure TClipboardPurger.Purge(aClipboard: PClipboard);
Var
  Prg : TPurgingRec;
  A : Array Of Char;
  I, Sum, Temp : DWORD;
  P : Pointer;
  h : THandle;
begin
  inherited;
  If Assigned(OnPurgeStart) Then
    OnPurgeStart(Self);
  Purging := True;
  SetPurgeProcedure;
  With Prg Do
  Begin
    Position := 0;
    Phase := 0;
    Length := 0;
    Finished := False;
  End;
  aClipboard^.Open;
  If (aClipboard^.FormatCount > 0) Then
  Begin
    For I := 1 to aClipboard^.FormatCount -1 Do
    Begin
      h := aClipboard^.GetAsHandle(aClipboard^.Formats[I]);
      P := GlobalLock(h);
      Inc(Prg.Length,Sizeof(P^));
      GlobalUnlock(h);
    End;
    SetLength(A,Prg.Length);
  End;
  If Prg.Length > 0 Then
    Repeat
      Sum := 0;
      PurgingProcedure(@Prg);
      If Prg.Position >= Prg.Length Then
      Begin
        Move(Prg.Buffer,A[Prg.Position-8], (8 -(Prg.Position - Prg.Length)) Mod 8);
        For I := 1 to aClipboard^.FormatCount -1 Do
        Begin
          h := aClipboard^.GetAsHandle(aClipboard^.Formats[I]);
          P := GlobalLock(h);
          Temp := Sizeof(P^);
          GetMem(P,Temp);
          Move(A[Sum],P^,Temp);
          Inc(Sum,Temp);
          GlobalUnlock(h);
          h := LocalHandle(P);
          aClipboard^.SetAsHandle(aClipboard^.Formats[I],h);
          FreeMem(P);
        End;
        Prg.Position := 0;
      End
      Else
        Move(Prg.Buffer,A[Prg.Position-8],8);
    Until Prg.Finished;
  aClipboard^.Close;
  aClipboard^.Clear;
  Purging := False;
  If Assigned(OnPurgeEnd) Then
    OnPurgeEnd(Self);
end;

procedure TClipboardPurger.PurgeTheClipboard;
Var
  Clip : TClipboard;
begin
  Clip := Clipboard;
  Purge(@Clip);
end;

{ TGlobalPurger }

procedure TGlobalPurger.AddToList(const NewItem: Pointer);
begin
  GlobalList.Add(NewItem);
end;

procedure TGlobalPurger.ClearList;
begin
  inherited;
  GlobalList.Clear;
  GlobalList.Capacity := 0;
end;

constructor TGlobalPurger.Create;
begin
  inherited;
  GlobalList := TList.Create;
end;

procedure TGlobalPurger.DeleteFromList(const ExstItem: Pointer);
begin
  If GlobalList.IndexOf(ExstItem) = -1 Then
    Raise EListError.Create('Item was not found on the list')
  Else
    GlobalList.Delete(GlobalList.IndexOf(ExstItem));
end;

destructor TGlobalPurger.Destroy;
begin
  GlobalList.Free;
  inherited;
end;

procedure TGlobalPurger.LoadList(List: TList);
begin
  GlobalList := List;
end;

procedure TGlobalPurger.MassPurge;
Var
  D : dWord;
begin
  inherited;
  If Assigned(OnMassPurgeStart) Then
    OnMassPurgeStart(Self);
  For D := 0 To GlobalList.Count -1 Do
  Begin
    Purge(GlobalList[D]);
    If FDeleteAfterPurge Then
    Begin
      DeleteFromList(GlobalList[D]);
    End;
  End;
  If Assigned(OnMassPurgeEnd) Then
    OnMassPurgeEnd(Self);
end;

procedure TGlobalPurger.Purge(Mem: Pointer);
Var
  Prg : TPurgingRec;
  A : Array Of Char;
  h : THandle;
begin
  inherited;
  If Assigned(OnPurgeStart) Then
    OnPurgeStart(Self);
  Purging := True;
  SetPurgeProcedure;
  Prg.Position := 0;
  Prg.Phase := 0;
  h := GlobalHandle(Mem);
  Prg.Length := GlobalSize(h);
  SetLength(A,Prg.Length);
  Prg.Finished := False;
  Repeat
    PurgingProcedure(@Prg);
    If Prg.Position >= Prg.Length Then
    Begin
      Move(Prg.Buffer,A[Prg.Position-8],(8 -(Prg.Position - Prg.Length)) Mod 8);
      CopyMemory(Mem,@A[0],Prg.Length);
      Prg.Position := 0;
    End
    Else
      Move(Prg.Buffer,A[Prg.Position-8],8);
  Until Prg.Finished;
  CloseHandle(h);
  Purging := False;
  If Assigned(OnPurgeEnd) Then
    OnPurgeEnd(Self);
End;

{ TLocalPurger }

procedure TLocalPurger.AddToList(const NewItem: Pointer);
begin
  LocalList.Add(NewItem);
end;

procedure TLocalPurger.ClearList;
begin
  inherited;
  LocalList.Clear;
end;

constructor TLocalPurger.Create;
begin
  inherited;
  LocalList := TList.Create;
end;

procedure TLocalPurger.DeleteFromList(const ExstItem: Pointer);
begin
  If LocalList.IndexOf(ExstItem) = -1 Then
    Raise EListError.Create('Item was not found on the list')
  Else
    LocalList.Delete(LocalList.IndexOf(ExstItem));
end;

destructor TLocalPurger.Destroy;
begin
  LocalList.Free;
  inherited;
end;

procedure TLocalPurger.LoadList(List: TList);
begin
  LocalList := List;
end;

procedure TLocalPurger.MassPurge;
Var
  D : dWord;
begin
  inherited;
  If Assigned(OnMassPurgeStart) Then
    OnMassPurgeStart(Self);
  For D := 0 To LocalList.Count -1 Do
  Begin
    Purge(LocalList[D]);
    If FDeleteAfterPurge Then
    Begin
      DeleteFromList(LocalList[D]);
    End;
  End;
  If Assigned(OnMassPurgeEnd) Then
    OnMassPurgeEnd(Self);
end;

procedure TLocalPurger.Purge(Mem: Pointer);
Var
  Prg : TPurgingRec;
  A : Array Of Char;
  h : THandle;
begin
  inherited;
  If Assigned(OnPurgeStart) Then
    OnPurgeStart(Self);
  Purging := True;
  SetPurgeProcedure;
  Prg.Position := 0;
  Prg.Phase := 0;
  h := LocalHandle(Mem);
  Prg.Length := LocalSize(h);
  SetLength(A,Prg.Length);
  Prg.Finished := False;
  Repeat
    PurgingProcedure(@Prg);
    If Prg.Position >= Prg.Length Then
    Begin
      Move(Prg.Buffer,A[Prg.Position-8],(8 -(Prg.Position - Prg.Length)) Mod 8);
      CopyMemory(Mem,@A[0],Prg.Length);
      Prg.Position := 0;
    End
    Else
      Move(Prg.Buffer,A[Prg.Position-8],8);
  Until Prg.Finished;
  CloseHandle(h);
  Purging := False;
  If Assigned(OnPurgeEnd) Then
    OnPurgeEnd(Self);
End;


end.
